"""dbt-exasol Adapter implementation extending SQLAdapter"""
from __future__ import absolute_import

from typing import Dict, Optional

import agate
from dbt.adapters.sql import SQLAdapter
from dbt.exceptions import raise_compiler_error
from dbt.utils import filter_null_values

from dbt.adapters.exasol import ExasolColumn, ExasolConnectionManager, ExasolRelation


class ExasolAdapter(SQLAdapter):
    """Exasol SQLAdapter extension"""

    Relation = ExasolRelation
    Column = ExasolColumn
    ConnectionManager = ExasolConnectionManager

    @classmethod
    def date_function(cls):
        return "current_timestamp()"

    @classmethod
    def is_cancelable(cls):
        return False

    @classmethod
    def convert_text_type(cls, agate_table, col_idx):
        return f"varchar({2000000})"

    def _make_match_kwargs(
        self, database: str, schema: str, identifier: str
    ) -> Dict[str, str]:
        quoting = self.config.quoting
        if identifier is not None and quoting["identifier"] is False:
            identifier = identifier.lower()

        if schema is not None and quoting["schema"] is False:
            schema = schema.lower()

        if database is not None and quoting["database"] is False:
            database = database.lower()

        return filter_null_values(
            {
                "identifier": identifier,
                "schema": schema,
            }
        )

    @classmethod
    def convert_number_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        decimals = agate_table.aggregate(agate.MaxPrecision(col_idx))
        return "float" if decimals else "integer"

    def timestamp_add_sql(
        self, add_to: str, number: int = 1, interval: str = "hour"
    ) -> str:
        """
        Overriding BaseAdapter default method because Exasol's syntax expects
        the number in quotes without the interval
        """
        return f"{add_to} + interval '{number}' {interval}"

    def quote_seed_column(self, column: str, quote_config: Optional[bool]) -> str:  # type: ignore
        quote_columns: bool = False
        if isinstance(quote_config, bool):
            quote_columns = quote_config
        elif quote_config is None:
            pass
        else:
            raise_compiler_error(
                f'The seed configuration value of "quote_columns" has an '
                f"invalid type {type(quote_config)}"
            )

        if quote_columns:
            return self.quote(column)
        return column



"""
testing sample
"""
import os

import pytest
from dbt.tests.util import run_dbt
from fixtures import my_model_sql, my_model_yml, my_seed_csv


# class must begin with 'Test'
class TestExample:
    """
    Methods in this class will be of two types:
    1. Fixtures defining the dbt "project" for this test case.
       These are scoped to the class, and reused for all tests in the class.
    2. Actual tests, whose names begin with 'test_'.
       These define sequences of dbt commands and 'assert' statements.
    """

    # configuration in dbt_project.yml
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"name": "example", "models": {"+materialized": "view"}}

    # everything that goes in the "seeds" directory
    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "my_seed.csv": my_seed_csv,
        }

    # everything that goes in the "models" directory
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model.sql": my_model_sql,
            "my_model.yml": my_model_yml,
        }

    # The actual sequence of dbt commands and assertions
    # pytest will take care of all "setup" + "teardown"
    def test_run_seed_test(self, project):
        """
        Seed, then run, then test. We expect one of the tests to fail
        An alternative pattern is to use pytest "xfail" (see below)
        """
        # seed seeds
        results = run_dbt(["seed"])
        assert len(results) == 1
        # run models
        results = run_dbt(["run"])
        assert len(results) == 1
        # test tests
        results = run_dbt(["test"], expect_pass=False)  # expect failing test
        assert len(results) == 2
        # validate that the results include one pass and one failure
        result_statuses = sorted(r.status for r in results)
        assert result_statuses == ["fail", "pass"]

    @pytest.mark.xfail
    def test_build(self, project):
        """Expect a failing test"""
        # do it all
        results = run_dbt(["build"])



"""dbt-exasol adapter relation module"""
from dataclasses import dataclass
from typing import Optional, Type, TypeVar

from dbt.adapters.base.relation import BaseRelation, Policy
from dbt.contracts.relation import RelationType


@dataclass
class ExasolQuotePolicy(Policy):
    """quote policy - not quotes"""

    database: bool = False
    schema: bool = False
    identifier: bool = False


Self = TypeVar("Self", bound="BaseRelation")


@dataclass(frozen=True, eq=False, repr=False)
class ExasolRelation(BaseRelation):
    """Relation implementation for exasol"""

    quote_policy: ExasolQuotePolicy = ExasolQuotePolicy()

    @classmethod
    # pylint: disable=too-many-arguments
    # pylint: disable=redefined-builtin
    # pylint: disable=unused-argument
    def create(
        cls: Type[Self],
        database: Optional[str] = None,
        schema: Optional[str] = None,
        identifier: Optional[str] = None,
        type: Optional[RelationType] = None,
        quote_policy: Type[ExasolQuotePolicy] = quote_policy,
        **kwargs,
    ) -> Self:
        """updating kwargs to set schema and identifier
        Args:
            cls (Type[Self]): _description_
            database (Optional[str], optional): _description_. Defaults to None.
            schema (Optional[str], optional): _description_. Defaults to None.
            identifier (Optional[str], optional): _description_. Defaults to None.
            type (Optional[RelationType], optional): _description_. Defaults to None.

        Returns:
            Self: _description_
        """
        kwargs.update(
            {
                "path": {
                    "schema": schema,
                    "identifier": identifier,
                },
                "type": type,
            }
        )
        return cls.from_dict(kwargs)

    @staticmethod
    def add_ephemeral_prefix(name: str):
        return f"dbt__CTE__{name}"



"""pytest conftest to load base dbt project fixtures
"""
import os

import pytest
from dotenv import load_dotenv

load_dotenv()

pytest_plugins = ["dbt.tests.fixtures.project"]


@pytest.fixture(scope="class")
def dbt_profile_target():
    """Fixture overwrite from dbt-core to configure profiles.yml

    Returns:
        dict: key values for profiles.yml
    """
    return {
        "type": "exasol",
        "threads": 1,
        "dsn": os.getenv("DBT_DSN", "localhost:8563"),
        "user": os.getenv("DBT_USER", "sys"),
        "pass": os.getenv("DBT_PASS", "exasol"),
        "dbname": "DB",
    }



from dbt.adapters.exasol.connections import ExasolConnectionManager
from dbt.adapters.exasol.connections import ExasolCredentials
from dbt.adapters.exasol.column import ExasolColumn
from dbt.adapters.exasol.relation import ExasolRelation
from dbt.adapters.exasol.impl import ExasolAdapter

from dbt.adapters.base import AdapterPlugin
from dbt.include import exasol


Plugin = AdapterPlugin(
    adapter=ExasolAdapter,
    credentials=ExasolCredentials,
    include_path=exasol.PACKAGE_PATH,
)



import os

import pytest
from dbt.tests.adapter.basic.files import config_materialized_view
from dbt.tests.util import run_dbt


@pytest.fixture(scope="class")
def dbt_profile_target():
    return {
        "type": "exasol",
        "threads": 1,
        "dsn": os.getenv("DBT_DSN"),
        "user": os.getenv("DBT_USER"),
        "pass": os.getenv("DBT_PASS"),
        "dbname": "DB",
        "timestamp_format": "YYYY-MM-DD HH:MI:SS",
    }


base_view_sql = config_materialized_view + "select 1 as id;"

seeds_tsspace_csv = """
id,name,some_date
1,Easton,1981-05-20 06:46:51
2,Lillian,1978-09-03 18:10:33
3,Jeremiah,1982-03-11 03:59:51
4,Nolan,1976-05-06 20:21:35
5,Hannah,1982-06-23 05:41:26
6,Eleanor,1991-08-10 23:12:21
7,Lily,1971-03-29 14:58:02
8,Jonathan,1988-02-26 02:55:24
9,Adrian,1994-02-09 13:14:23
10,Nora,1976-03-01 16:51:39
""".lstrip()


class TestViewModelSemicolon:
    @pytest.fixture(scope="class")
    def models(self):
        return {"view_model.sql": base_view_sql}

    def test_view(self, project):
        results = run_dbt(["build"], expect_pass=False)
        assert len(results) == 1
        assert results[0].status == "error"


class TestCustomTimestampFormat:
    @pytest.fixture(scope="class")
    def seeds(self):
        return {"ts_space.csv": seeds_tsspace_csv}

    def test_custom_ts_format(self, project):

        results = run_dbt(["seed"])
        assert len(results) == 1
        assert results[0].status == "success"



import os

import pytest
from dbt.exceptions import CompilationException
from dbt.tests.adapter.utils.base_utils import BaseUtils

# from dbt.tests.adapter.utils.test_array_append import BaseArrayAppend
# from dbt.tests.adapter.utils.test_array_concat import BaseArrayConcat
# from dbt.tests.adapter.utils.test_array_construct import BaseArrayConstruct
from dbt.tests.adapter.utils.test_any_value import BaseAnyValue
from dbt.tests.adapter.utils.test_bool_or import BaseBoolOr
from dbt.tests.adapter.utils.test_cast_bool_to_text import BaseCastBoolToText
from dbt.tests.adapter.utils.test_concat import BaseConcat
from dbt.tests.adapter.utils.test_date_trunc import BaseDateTrunc

# from dbt.tests.adapter.utils.test_current_timestamp import BaseCurrentTimestampAware
from dbt.tests.adapter.utils.test_dateadd import BaseDateAdd
from dbt.tests.adapter.utils.test_datediff import BaseDateDiff
from dbt.tests.adapter.utils.test_escape_single_quotes import (
    BaseEscapeSingleQuotesBackslash,
    BaseEscapeSingleQuotesQuote,
)
from dbt.tests.adapter.utils.test_except import BaseExcept
from dbt.tests.adapter.utils.test_hash import BaseHash
from dbt.tests.adapter.utils.test_intersect import BaseIntersect
from dbt.tests.adapter.utils.test_last_day import BaseLastDay
from dbt.tests.adapter.utils.test_length import BaseLength
from dbt.tests.adapter.utils.test_listagg import BaseListagg
from dbt.tests.adapter.utils.test_position import BasePosition
from dbt.tests.adapter.utils.test_replace import BaseReplace
from dbt.tests.adapter.utils.test_right import BaseRight
from dbt.tests.adapter.utils.test_safe_cast import BaseSafeCast
from dbt.tests.adapter.utils.test_split_part import BaseSplitPart
from dbt.tests.adapter.utils.test_string_literal import BaseStringLiteral
from dbt.tests.util import run_dbt
from utils_fixtures import *


class TestAnyValue(BaseAnyValue):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_any_value.yml": exasol__models__test_any_value_yml,
            "test_any_value.sql": self.interpolate_macro_namespace(
                exasol__models__test_any_value_sql, "any_value"
            ),
        }


# class TestArrayAppend(BaseArrayAppend):
#    pass


# class TestArrayConcat(BaseArrayConcat):
#    pass


# class TestArrayConstruct(BaseArrayConstruct):
#    pass


class TestBoolOr(BaseBoolOr):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "data_bool_or.csv": exasol__seeds__data_bool_or_csv,
            "data_bool_or_expected.csv": exasol__seeds__data_bool_or_expected_csv,
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_bool_or.yml": exasol__models__test_bool_or_yml,
            "test_bool_or.sql": self.interpolate_macro_namespace(
                exasol__models__test_bool_or_sql, "bool_or"
            ),
        }


class TestCastBoolToText(BaseCastBoolToText):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_cast_bool_to_text.yml": exasol__models__test_cast_bool_to_text_yml,
            "test_cast_bool_to_text.sql": self.interpolate_macro_namespace(
                exasol__models__test_cast_bool_to_text_sql, "cast_bool_to_text"
            ),
        }


class TestConcat(BaseConcat):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {"data_concat.csv": exasol__seeds__data_concat_csv}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_concat.yml": exasol__models__test_concat_yml,
            "test_concat.sql": self.interpolate_macro_namespace(
                exasol__models__test_concat_sql, "concat"
            ),
        }


# Use either BaseCurrentTimestampAware or BaseCurrentTimestampNaive but not both
# class TestCurrentTimestamp(BaseCurrentTimestampAware):
#    pass


class TestDateAdd(BaseDateAdd):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "name": "test",
            # this is only needed for BigQuery, right?
            # no harm having it here until/unless there's an adapter that doesn't support the 'timestamp' type
            "seeds": {
                "test": {
                    "data_dateadd": {
                        "+column_types": {
                            "from_time": "timestamp",
                            "result": "timestamp",
                        },
                    },
                },
            },
        }

    @pytest.fixture(scope="class")
    def seeds(self):
        return {"data_dateadd.csv": exasol__seeds__data_dateadd_csv}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_dateadd.yml": exasol__models__test_dateadd_yml,
            "test_dateadd.sql": self.interpolate_macro_namespace(
                exasol__models__test_dateadd_sql, "dateadd"
            ),
        }


class TestDateDiff(BaseDateDiff):
    @pytest.fixture(scope="class")
    def dbt_profile_target(self):
        return {
            "type": "exasol",
            "threads": 1,
            "dsn": os.getenv("DBT_DSN", "localhost:8563"),
            "user": os.getenv("DBT_USER", "sys"),
            "pass": os.getenv("DBT_PASS", "exasol"),
            "dbname": "DB",
            "timestamp_format": "YYYY-MM-DD HH:MI:SS.FF6",
        }

    @pytest.fixture(scope="class")
    def seeds(self):
        return {"data_datediff.csv": exasol__seeds__data_datediff_csv}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_datediff.yml": exasol__models__test_datediff_yml,
            "test_datediff.sql": self.interpolate_macro_namespace(
                exasol__models__test_datediff_sql, "datediff"
            ),
        }


class TestDateTrunc(BaseDateTrunc):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {"data_date_trunc.csv": exasol__seeds__data_date_trunc_csv}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_date_trunc.yml": exasol__models__test_date_trunc_yml,
            "test_date_trunc.sql": self.interpolate_macro_namespace(
                exasol__models__test_date_trunc_sql, "date_trunc"
            ),
        }


class TestEscapeSingleQuotes(BaseEscapeSingleQuotesBackslash):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_escape_single_quotes.yml": exasol__models__test_escape_single_quotes_yml,
            "test_escape_single_quotes.sql": self.interpolate_macro_namespace(
                exasol__models__test_escape_single_quotes_quote_sql,
                "escape_single_quotes",
            ),
        }


class BaseEscapeSingleQuotesBackslash(BaseUtils):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_escape_single_quotes.yml": exasol__models__test_escape_single_quotes_yml,
            "test_escape_single_quotes.sql": self.interpolate_macro_namespace(
                exasol__models__test_escape_single_quotes_backslash_sql,
                "escape_single_quotes",
            ),
        }


class TestExcept(BaseExcept):
    pass


class TestHash(BaseHash):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {"data_hash.csv": exasol__seeds__data_hash_csv}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_hash.yml": exasol__models__test_hash_yml,
            "test_hash.sql": self.interpolate_macro_namespace(
                exasol__models__test_hash_sql, "hash"
            ),
        }


class TestIntersect(BaseIntersect):
    pass


class TestLastDay(BaseLastDay):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {"data_last_day.csv": exasol__seeds__data_last_day_csv}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_last_day.yml": exasol__models__test_last_day_yml,
            "test_last_day.sql": self.interpolate_macro_namespace(
                exasol__models__test_last_day_sql, "last_day"
            ),
        }


class TestLength(BaseLength):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {"data_length.csv": exasol__seeds__data_length_csv}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_length.yml": exasol__models__test_length_yml,
            "test_length.sql": self.interpolate_macro_namespace(
                exasol__models__test_length_sql, "length"
            ),
        }


class TestListagg(BaseListagg):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "data_listagg.csv": exasol__seeds__data_listagg_csv,
            "data_listagg_output.csv": exasol__seeds__data_listagg_output_csv,
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_listagg.yml": exasol__models__test_listagg_yml,
            "test_listagg.sql": self.interpolate_macro_namespace(
                exasol__models__test_listagg_sql, "listagg"
            ),
        }

    def test_build_assert_equal(self, project):
        with pytest.raises(CompilationException) as exc_info:
            run_dbt(["build"], expect_pass=False)
        assert exc_info.value.msg == "`limit_num` parameter is not supported on Exasol!"


class TestPosition(BasePosition):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {"data_position.csv": exasol__seeds__data_position_csv}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_position.yml": exasol__models__test_position_yml,
            "test_position.sql": self.interpolate_macro_namespace(
                exasol__models__test_position_sql, "position"
            ),
        }


class TestReplace(BaseReplace):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {"data_replace.csv": exasol__seeds__data_replace_csv}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_replace.yml": exasol__models__test_replace_yml,
            "test_replace.sql": self.interpolate_macro_namespace(
                exasol__models__test_replace_sql, "replace"
            ),
        }


class TestRight(BaseRight):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {"data_right.csv": exasol__seeds__data_right_csv}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_right.yml": exasol__models__test_right_yml,
            "test_right.sql": self.interpolate_macro_namespace(
                exasol__models__test_right_sql, "right"
            ),
        }


class TestSafeCast(BaseSafeCast):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {"data_safe_cast.csv": exasol__seeds__data_safe_cast_csv}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_safe_cast.yml": exasol__models__test_safe_cast_yml,
            "test_safe_cast.sql": self.interpolate_macro_namespace(
                self.interpolate_macro_namespace(
                    exasol__models__test_safe_cast_sql, "safe_cast"
                ),
                "type_string",
            ),
        }


class TestSplitPart(BaseSplitPart):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {"data_split_part.csv": exasol__seeds__data_split_part_csv}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_split_part.yml": exasol__models__test_split_part_yml,
            "test_split_part.sql": self.interpolate_macro_namespace(
                exasol__models__test_split_part_sql, "split_part"
            ),
        }

    def test_build_assert_equal(self, project):
        with pytest.raises(CompilationException) as exc_info:
            run_dbt(["build"], expect_pass=False)
        assert exc_info.value.msg == "Unsupported on Exasol! Sorry..."


class TestStringLiteral(BaseStringLiteral):
    pass



import pytest, os

from dbt.tests.adapter.basic.test_adapter_methods import BaseAdapterMethod
from dbt.tests.adapter.basic.test_base import BaseSimpleMaterializations
from dbt.tests.adapter.basic.test_empty import BaseEmpty
from dbt.tests.adapter.basic.test_ephemeral import BaseEphemeral
from dbt.tests.adapter.basic.test_generic_tests import BaseGenericTests
from dbt.tests.adapter.basic.test_incremental import BaseIncremental
from dbt.tests.adapter.basic.test_singular_tests import BaseSingularTests
from dbt.tests.adapter.basic.test_singular_tests_ephemeral import (
    BaseSingularTestsEphemeral,
)
from dbt.tests.adapter.basic.test_snapshot_check_cols import BaseSnapshotCheckCols
from dbt.tests.adapter.basic.test_snapshot_timestamp import BaseSnapshotTimestamp
from dbt.tests.adapter.basic.test_validate_connection import BaseValidateConnection
from test_docs_generate import (BaseDocsGenerate, BaseDocsGenReferences)


class TestSimpleMaterializationsExasol(BaseSimpleMaterializations):
    pass


class TestSingularTestsExasol(BaseSingularTests):
    pass


class TestSingularTestsEphemeralExasol(BaseSingularTestsEphemeral):
    pass


class TestEmptyExasol(BaseEmpty):
    pass


class TestEphemeralExasol(BaseEphemeral):
    pass


class TestIncrementalExasol(BaseIncremental):
    pass


class TestGenericTestsExasol(BaseGenericTests):
    pass


class TestSnapshotCheckColsExasol(BaseSnapshotCheckCols):
    pass


class TestSnapshotTimestampExasol(BaseSnapshotTimestamp):
    pass


class TestBaseAdapterMethod(BaseAdapterMethod):
    pass


class TestBaseCachingExasol(BaseAdapterMethod):
    pass


class TestValidateConnectionExasol(BaseValidateConnection):
    pass

class TestDocsGenerateExasol(BaseDocsGenerate):
    @pytest.fixture(scope="class")
    def dbt_profile_target(self):
        return {
           "type": "exasol",
           "threads": 1,
           "dsn": os.getenv('DBT_DSN',"localhost:8563"),
           "user": os.getenv('DBT_USER',"sys"),
           "pass": os.getenv('DBT_PASS',"exasol"),
           "dbname": "DB",
           "timestamp_format": "YYYY-MM-DD HH:MI:SS.FF6"
        }

class TestDocsGenReferencesExasol(BaseDocsGenReferences):
    
    @pytest.fixture(scope="class")
    def dbt_profile_target(self):
        return {
           "type": "exasol",
           "threads": 1,
           "dsn": os.getenv('DBT_DSN',"localhost:8563"),
           "user": os.getenv('DBT_USER',"sys"),
           "pass": os.getenv('DBT_PASS',"exasol"),
           "dbname": "DB",
           "timestamp_format": "YYYY-MM-DD HH:MI:SS.FF6"
        }


import pytest, os
#from dbt.tests.adapter.utils.data_types.test_type_bigint import BaseTypeBigInt
from dbt.tests.adapter.utils.data_types.test_type_float import BaseTypeFloat
from dbt.tests.adapter.utils.data_types.test_type_int import BaseTypeInt
from dbt.tests.adapter.utils.data_types.test_type_numeric import BaseTypeNumeric
from dbt.tests.adapter.utils.data_types.test_type_string import BaseTypeString
from dbt.tests.adapter.utils.data_types.test_type_timestamp import BaseTypeTimestamp
from test_type_bigint import BaseTypeBigInt
#from dbt.tests.adapter.utils.data_types.test_type_boolean import BaseTypeBoolean


class TestTypeBigInt(BaseTypeBigInt):
    pass

    
class TestTypeFloat(BaseTypeFloat):
    pass

    
class TestTypeInt(BaseTypeInt):
    pass

    
class TestTypeNumeric(BaseTypeNumeric):
    pass

    
class TestTypeString(BaseTypeString):
    pass

    
class TestTypeTimestamp(BaseTypeTimestamp):
    @pytest.fixture(scope="class")
    def dbt_profile_target(self):
        return {
           "type": "exasol",
           "threads": 1,
           "dsn": os.getenv('DBT_DSN',"localhost:8563"),
           "user": os.getenv('DBT_USER',"sys"),
           "pass": os.getenv('DBT_PASS',"exasol"),
           "dbname": "DB",
           "timestamp_format": "YYYY-MM-DD HH:MI:SS.FF6"
        }


#class TestTypeBoolean(BaseTypeBoolean):
#    pass


import pytest
from dbt.tests.adapter.incremental.test_incremental_unique_id import (
    BaseIncrementalUniqueKey,
)

models__trinary_unique_key_list_sql = """
-- a multi-argument unique key list should see overwriting on rows in the model
--   where all unique key fields apply

{{
    config(
        materialized='incremental',
        unique_key=['states', 'county', 'city']
    )
}}

select
    states as states,
    county as county,
    city as city,
    last_visit_date as last_visit_date
from {{ ref('seed') }}

{% if is_incremental() %}
    where last_visit_date > (select max(last_visit_date) from {{ this }})
{% endif %}

"""

models__nontyped_trinary_unique_key_list_sql = """
-- a multi-argument unique key list should see overwriting on rows in the model
--   where all unique key fields apply
--   N.B. needed for direct comparison with seed

{{
    config(
        materialized='incremental',
        unique_key=['states', 'county', 'city']
    )
}}

select
    states as states,
    county as county,
    city as city,
    last_visit_date as last_visit_date
from {{ ref('seed') }}

{% if is_incremental() %}
    where last_visit_date > (select max(last_visit_date) from {{ this }})
{% endif %}

"""

models__unary_unique_key_list_sql = """
-- a one argument unique key list should result in overwritting semantics for
--   that one matching field

{{
    config(
        materialized='incremental',
        unique_key=['states']
    )
}}

select
    states as states,
    county as county,
    city as city,
    last_visit_date as last_visit_date
from {{ ref('seed') }}

{% if is_incremental() %}
    where last_visit_date > (select max(last_visit_date) from {{ this }})
{% endif %}

"""

models__not_found_unique_key_sql = """
-- a model with a unique key not found in the table itself will error out

{{
    config(
        materialized='incremental',
        unique_key='thisisnotacolumn'
    )
}}

select
    *
from {{ ref('seed') }}

{% if is_incremental() %}
    where last_visit_date > (select max(last_visit_date) from {{ this }})
{% endif %}

"""

models__empty_unique_key_list_sql = """
-- model with empty list unique key should build normally

{{
    config(
        materialized='incremental',
        unique_key=[]
    )
}}

select * from {{ ref('seed') }}

{% if is_incremental() %}
    where last_visit_date > (select max(last_visit_date) from {{ this }})
{% endif %}

"""

models__no_unique_key_sql = """
-- no specified unique key should cause no special build behavior

{{
    config(
        materialized='incremental'
    )
}}

select
    *
from {{ ref('seed') }}

{% if is_incremental() %}
    where last_visit_date > (select max(last_visit_date) from {{ this }})
{% endif %}

"""

models__empty_str_unique_key_sql = """
-- ensure model with empty string unique key should build normally

{{
    config(
        materialized='incremental',
        unique_key=''
    )
}}

select
    *
from {{ ref('seed') }}

{% if is_incremental() %}
    where last_visit_date > (select max(last_visit_date) from {{ this }})
{% endif %}

"""

models__str_unique_key_sql = """
-- a unique key with a string should trigger to overwrite behavior when
--   the source has entries in conflict (i.e. more than one row per unique key
--   combination)

{{
    config(
        materialized='incremental',
        unique_key='states'
    )
}}

select
    states as states,
    county as county,
    city as city,
    last_visit_date as last_visit_date
from {{ ref('seed') }}

{% if is_incremental() %}
    where last_visit_date > (select max(last_visit_date) from {{ this }})
{% endif %}

"""

models__duplicated_unary_unique_key_list_sql = """
{{
    config(
        materialized='incremental',
        unique_key=['states', 'states']
    )
}}

select
    states as states,
    county as county,
    city as city,
    last_visit_date as last_visit_date
from {{ ref('seed') }}

{% if is_incremental() %}
    where last_visit_date > (select max(last_visit_date) from {{ this }})
{% endif %}

"""

models__not_found_unique_key_list_sql = """
-- a unique key list with any element not in the model itself should error out

{{
    config(
        materialized='incremental',
        unique_key=['state', 'thisisnotacolumn']
    )
}}

select * from {{ ref('seed') }}

"""

models__expected__one_str__overwrite_sql = """
{{
    config(
        materialized='table'
    )
}}

select
    'CT' as states,
    'Hartford' as county,
    'Hartford' as city,
    cast('2022-02-14' as date) as last_visit_date
union all
select 'MA','Suffolk','Boston',cast('2020-02-12' as date)
union all
select 'NJ','Mercer','Trenton',cast('2022-01-01' as date)
union all
select 'NY','Kings','Brooklyn',cast('2021-04-02' as date)
union all
select 'NY','New York','Manhattan',cast('2021-04-01' as date)
union all
select 'PA','Philadelphia','Philadelphia',cast('2021-05-21' as date)

"""

models__expected__unique_key_list__inplace_overwrite_sql = """
{{
    config(
        materialized='table'
    )
}}

select
    'CT' as states,
    'Hartford' as county,
    'Hartford' as city,
    cast('2022-02-14' as date) as last_visit_date
union all
select 'MA','Suffolk','Boston',cast('2020-02-12' as date)
union all
select 'NJ','Mercer','Trenton',cast('2022-01-01' as date)
union all
select 'NY','Kings','Brooklyn',cast('2021-04-02' as date)
union all
select 'NY','New York','Manhattan',cast('2021-04-01' as date)
union all
select 'PA','Philadelphia','Philadelphia',cast('2021-05-21' as date)

"""

seeds__duplicate_insert_sql = """
-- Insert statement which when applied to seed.csv triggers the inplace
--   overwrite strategy of incremental models. Seed and incremental model
--   diverge.

-- insert new row, which should not be in incremental model
--  with primary or first three columns unique
insert into {schema}.seed
    (states, county, city, last_visit_date)
values ('CT','Hartford','Hartford','2022-02-14');

"""

seeds__seed_csv = """states,county,city,last_visit_date
CT,Hartford,Hartford,2020-09-23
MA,Suffolk,Boston,2020-02-12
NJ,Mercer,Trenton,2022-01-01
NY,Kings,Brooklyn,2021-04-02
NY,New York,Manhattan,2021-04-01
PA,Philadelphia,Philadelphia,2021-05-21
"""

seeds__add_new_rows_sql = """
-- Insert statement which when applied to seed.csv sees incremental model
--   grow in size while not (necessarily) diverging from the seed itself.

-- insert two new rows, both of which should be in incremental model
--   with any unique columns
insert into {schema}.seed
    (states, county, city, last_visit_date)
values ('WA','King','Seattle','2022-02-01');

insert into {schema}.seed
    (states, county, city, last_visit_date)
values ('CA','Los Angeles','Los Angeles','2022-02-01');

"""


class TestIncrementalUniqueKey(BaseIncrementalUniqueKey):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "trinary_unique_key_list.sql": models__trinary_unique_key_list_sql,
            "nontyped_trinary_unique_key_list.sql": models__nontyped_trinary_unique_key_list_sql,
            "unary_unique_key_list.sql": models__unary_unique_key_list_sql,
            "not_found_unique_key.sql": models__not_found_unique_key_sql,
            "empty_unique_key_list.sql": models__empty_unique_key_list_sql,
            "no_unique_key.sql": models__no_unique_key_sql,
            "empty_str_unique_key.sql": models__empty_str_unique_key_sql,
            "str_unique_key.sql": models__str_unique_key_sql,
            "duplicated_unary_unique_key_list.sql": models__duplicated_unary_unique_key_list_sql,
            "not_found_unique_key_list.sql": models__not_found_unique_key_list_sql,
            "expected": {
                "one_str__overwrite.sql": models__expected__one_str__overwrite_sql,
                "unique_key_list__inplace_overwrite.sql": models__expected__unique_key_list__inplace_overwrite_sql,
            },
        }

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "duplicate_insert.sql": seeds__duplicate_insert_sql,
            "seed.csv": seeds__seed_csv,
            "add_new_rows.sql": seeds__add_new_rows_sql,
        }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dbt',
 'dbt.adapters',
 'dbt.adapters.exasol',
 'dbt.include',
 'dbt.include.exasol',
 'tests',
 'tests.functional',
 'tests.functional.adapter',
 'tests.functional.adapter.utils',
 'tests.functional.adapter.utils.data_types']

package_data = \
{'': ['*'],
 'dbt.include.exasol': ['macros/*',
                        'macros/materializations/*',
                        'macros/utils/*']}

install_requires = \
['dbt-core>=1.2,<1.3',
 'dbt-tests-adapter>=1.2,<1.3',
 'pyexasol>=0.25.0,<0.26.0']

setup_kwargs = {
    'name': 'dbt-exasol',
    'version': '1.2.2',
    'description': 'Adapter to dbt-core for warehouse Exasol',
    'long_description': "# dbt-exasol\n**[dbt](https://www.getdbt.com/)** enables data analysts and engineers to transform their data using the same practices that software engineers use to build applications.\n\nPlease see the dbt documentation on **[Exasol setup](https://docs.getdbt.com/reference/warehouse-setups/exasol-setup)** for more information on how to start using the Exasol adapter.\n\n# Current profile.yml settings\n<File name='profiles.yml'>\n\n```yaml\ndbt-exasol:\n  target: dev\n  outputs:\n    dev:\n      type: exasol\n      threads: 1\n      dsn: HOST:PORT\n      user: USERNAME\n      password: PASSWORD\n      dbname: db\n      schema: SCHEMA\n```\n\n#### Optional parameters\n<ul>\n  <li><strong>connection_timeout</strong>: defaults to pyexasol default</li>\n  <li><strong>socket_timeout</strong>: defaults to pyexasol default</li>\n  <li><strong>query_timeout</strong>: defaults to pyexasol default</li>\n  <li><strong>compression</strong>: default: False</li>\n  <li><strong>encryption</strong>: default: False</li>\n  <li><strong>protocol_version</strong>: default: v3</li>\n  <li><strong>row_separator</strong>: default: CRLF for windows - LF otherwise</li>\n  <li><strong>timestamp_format</strong>: default: YYYY-MM-DDTHH:MI:SS.FF6</li>\n</ul>\n\n# Known isues\n\n## Breaking changes with release 1.2.2\n- Timestamp format defaults to YYYY-MM-DDTHH:MI:SS.FF6\n\n## SQL functions compatibility\n\n### split_part\nThere is no equivalent SQL function in Exasol for split_part.\n\n### listagg part_num\nThe SQL function listagg in Exasol does not support the num_part parameter.\n\n## Utilities shim package\nIn order to support packages like dbt-utils and dbt-audit-helper, we needed to create the [shim package exasol-utils](https://github.com/tglunde/exasol-utils). In this shim package we need to adapt to parts of the SQL functionality that is not compatible with Exasol - e.g. when 'final' is being used which is a keyword in Exasol. Please visit [Adaopter dispatch documentation](https://docs.getdbt.com/guides/advanced/adapter-development/3-building-a-new-adapter#adapter-dispatch) of dbt-labs for more information. \n# Reporting bugs and contributing code\n- Please report bugs using the issues\n\n# Release History\n\n## Release 1.2.2\n- Added timestamp format parameter in profile.yml parameter file to set Exasol session parameter NLS_TIMESTAMP_FORMAT when opening a connection. Defaults to 'YYYY-MM-DDTHH:MI:SS.FF6' \n- Adding row_separator (LF/CRLF) parameter in profile.yml parameter file to be used in seed csv import. Defaults to operating system default (os.linesep in python).\n- bugfix #36 regarding column quotes and case sensitivity of column names. \n- bugfix #42 regarding datatype change when using snapshot materialization. Added modify column statement in exasol__alter_column_type macro\n- bugfix #17 number format datatype\n- issue #24 - dbt-core v1.2.0 compatibility finished\n\n## Release 1.2.0\n- support for invalidate_hard_deletes option in snapshots added by jups23\n- added persist_docs support by sti0\n- added additional configuration keys that can be included in profiles.yml by johannes-becker-otto\n- added cross-database macros introduced in 1.2 by sti0\n- added support for connection retries by sti0\n- added support for grants by sti0\n- added pytest functional adapter tests by tglunde\n- tox testing for python 3.7.2 through 3.10 added by tglunde\n \n## Release 1.0.0\n- pyexasol HTTP import csv feature implemented. Optimal performance and compatibility with Exasol CSV parsing\n",
    'author': 'Torsten Glunde',
    'author_email': 'torsten.glunde@alligator-company.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://alligatorcompany.gitlab.io/dbt-exasol',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)



import pytest
from dbt.tests.adapter.grants.test_incremental_grants import BaseIncrementalGrants
from dbt.tests.adapter.grants.test_invalid_grants import BaseInvalidGrants
from dbt.tests.adapter.grants.test_model_grants import BaseModelGrants
from dbt.tests.adapter.grants.test_seed_grants import BaseSeedGrants
from dbt.tests.adapter.grants.test_snapshot_grants import BaseSnapshotGrants


class TestIncrementalGrantsExasol(BaseIncrementalGrants):
    pass


class TestInvalidGrantsExasol(BaseInvalidGrants):
    def grantee_does_not_exist_error(self):
        return "user or role 'INVALID_USER' does not exist"

    def privilege_does_not_exist_error(self):
        return "syntax error, unexpected ON_, expecting ',' or TO_ [line 3, column 34]"


class TestBaseModelGrantsExasol(BaseModelGrants):
    pass


class TestBaseSeedGrantsExasol(BaseSeedGrants):
    pass


class TestSnapshotGrantsExasol(BaseSnapshotGrants):
    pass



# any_value

exasol__models__test_any_value_sql = """
with testdata as (

    select * from {{ ref('data_any_value') }}

),

data_output as (

    select * from {{ ref('data_any_value_expected') }}

),

calculate as (
    select
        key_name,
        {{ any_value('static_col') }} as static_col,
        count(id) as num_rows
    from testdata
    group by key_name
)

select
    calculate.num_rows as actual,
    data_output.num_rows as expected
from calculate
left join data_output
on calculate.key_name = data_output.key_name
and calculate.static_col = data_output.static_col
"""


exasol__models__test_any_value_yml = """
version: 2
models:
  - name: test_any_value
    tests:
      - assert_equal:
          actual: actual
          expected: expected
"""
# bool_or

exasol__seeds__data_bool_or_csv = """key_column,val1,val2
abc,1,1
abc,1,0
def,1,0
hij,1,1
hij,1,
klm,1,0
klm,1,
"""


exasol__seeds__data_bool_or_expected_csv = """key_column,val
abc,true
def,false
hij,true
klm,false
"""


exasol__models__test_bool_or_sql = """
with testdata as (

    select * from {{ ref('data_bool_or') }}

),

data_output as (

    select * from {{ ref('data_bool_or_expected') }}

),

calculate as (

    select
        key_column,
        {{ bool_or('val1 = val2') }} as val
    from testdata
    group by key_column

)

select
    calculate.val as actual,
    data_output.val as expected
from calculate
left join data_output
on calculate.key_column = data_output.key_column
"""


exasol__models__test_bool_or_yml = """
version: 2
models:
  - name: test_bool_or
    tests:
      - assert_equal:
          actual: actual
          expected: expected
"""

# cast_bool_to_text
exasol__models__test_cast_bool_to_text_sql = """
with testdata as (

    select 0=1 as inputval, 'FALSE' as expected union all
    select 1=1 as inputval, 'TRUE' as expected union all
    select null as inputval, null as expected

)

select

    {{ cast_bool_to_text("inputval") }} as actual,
    expected

from testdata
"""

exasol__models__test_cast_bool_to_text_yml = """
version: 2
models:
  - name: test_cast_bool_to_text
    tests:
      - assert_equal:
          actual: actual
          expected: expected
"""

# concat

exasol__seeds__data_concat_csv = """input_1,input_2,output_val
a,b,ab
a,,a
,b,b
,,
"""


exasol__models__test_concat_sql = """
with testdata as (

    select * from {{ ref('data_concat') }}

)

select
    {{ concat(['input_1', 'input_2']) }} as actual,
    output_val as expected

from testdata
"""


exasol__models__test_concat_yml = """
version: 2
models:
  - name: test_concat
    tests:
      - assert_equal:
          actual: actual
          expected: expected
"""

# dateadd

exasol__seeds__data_dateadd_csv = """from_time,interval_length,datepart,res
2018-01-01T01:00:00,1,day,2018-01-02T01:00:00
2018-01-01T01:00:00,1,month,2018-02-01T01:00:00
2018-01-01T01:00:00,1,year,2019-01-01T01:00:00
2018-01-01T01:00:00,1,hour,2018-01-01T02:00:00
,1,day,
"""


exasol__models__test_dateadd_sql = """
with testdata as (

    select * from {{ ref('data_dateadd') }}

)

select
    case
        when datepart = 'hour' then cast({{ dateadd('hour', 'interval_length', 'from_time') }} as {{ api.Column.translate_type('timestamp') }})
        when datepart = 'day' then cast({{ dateadd('day', 'interval_length', 'from_time') }} as {{ api.Column.translate_type('timestamp') }})
        when datepart = 'month' then cast({{ dateadd('month', 'interval_length', 'from_time') }} as {{ api.Column.translate_type('timestamp') }})
        when datepart = 'year' then cast({{ dateadd('year', 'interval_length', 'from_time') }} as {{ api.Column.translate_type('timestamp') }})
        else null
    end as actual,
    res as expected

from testdata
"""

exasol__models__test_dateadd_yml = """
version: 2
models:
  - name: test_dateadd
    tests:
      - assert_equal:
          actual: actual
          expected: expected
"""

# datediff

exasol__seeds__data_datediff_csv = """first_date,second_date,datepart,res
2018-01-01 01:00:00,2018-01-02 01:00:00,day,1
2018-01-01 01:00:00,2018-02-01 01:00:00,month,1
2018-01-01 01:00:00,2019-01-01 01:00:00,year,1
2018-01-01 01:00:00,2018-01-01 02:00:00,hour,1
2018-01-01 01:00:00,2018-01-01 02:01:00,minute,61
2018-01-01 01:00:00,2018-01-01 02:00:01,second,3601
,2018-01-01 02:00:00,hour,
2018-01-01 02:00:00,,hour,
"""


exasol__models__test_datediff_sql = """
with testdata as (

    select * from {{ ref('data_datediff') }}

)

select

    case
        when datepart = 'second' then {{ datediff('first_date', 'second_date', 'second') }}
        when datepart = 'minute' then {{ datediff('first_date', 'second_date', 'minute') }}
        when datepart = 'hour' then {{ datediff('first_date', 'second_date', 'hour') }}
        when datepart = 'day' then {{ datediff('first_date', 'second_date', 'day') }}
        when datepart = 'month' then {{ datediff('first_date', 'second_date', 'month') }}
        when datepart = 'year' then {{ datediff('first_date', 'second_date', 'year') }}
        else null
    end as actual,
    res as expected

from testdata

-- Also test correct casting of literal values.

union all select {{ datediff("'1999-12-31 23:59:59.999999'", "'2000-01-01 00:00:00.000000'", 'second') }} as actual, 1 as expected
union all select {{ datediff("'1999-12-31 23:59:59.999999'", "'2000-01-01 00:00:00.000000'", 'minute') }} as actual, 1 as expected
union all select {{ datediff("'1999-12-31 23:59:59.999999'", "'2000-01-01 00:00:00.000000'", 'hour') }} as actual, 1 as expected
union all select {{ datediff("'1999-12-31 23:59:59.999999'", "'2000-01-01 00:00:00.000000'", 'day') }} as actual, 1 as expected
union all select {{ datediff("'1999-12-31 23:59:59.999999'", "'2000-01-01 00:00:00.000000'", 'month') }} as actual, 1 as expected
union all select {{ datediff("'1999-12-31 23:59:59.999999'", "'2000-01-01 00:00:00.000000'", 'year') }} as actual, 1 as expected
"""


exasol__models__test_datediff_yml = """
version: 2
models:
  - name: test_datediff
    tests:
      - assert_equal:
          actual: actual
          expected: expected
"""

# date_trunc

exasol__seeds__data_date_trunc_csv = """updated_at,day_val,month_val
2018-01-05T12:00:00,2018-01-05,2018-01-01
,,
"""


exasol__models__test_date_trunc_sql = """
with testdata as (

    select * from {{ ref('data_date_trunc') }}

)

select
    cast({{date_trunc('day', 'updated_at') }} as date) as actual,
    day_val as expected

from testdata

union all

select
    cast({{ date_trunc('month', 'updated_at') }} as date) as actual,
    month_val as expected

from testdata
"""


exasol__models__test_date_trunc_yml = """
version: 2
models:
  - name: test_date_trunc
    tests:
      - assert_equal:
          actual: actual
          expected: expected
"""

# escape_single_quotes

exasol__models__test_escape_single_quotes_quote_sql = """
select '{{ escape_single_quotes("they're") }}' as actual, 'they''re' as expected union all
select '{{ escape_single_quotes("they are") }}' as actual, 'they are' as expected
"""


# The expected literal is 'they\'re'. The second backslash is to escape it from Python.
exasol__models__test_escape_single_quotes_backslash_sql = """
select '{{ escape_single_quotes("they're") }}' as actual, 'they\'re' as expected union all
select '{{ escape_single_quotes("they are") }}' as actual, 'they are' as expected
"""


exasol__models__test_escape_single_quotes_yml = """
version: 2
models:
  - name: test_escape_single_quotes
    tests:
      - assert_equal:
          actual: actual
          expected: expected
"""


# hash

exasol__seeds__data_hash_csv = """input_1,output_val
ab,187ef4436122d1cc2f40dc2b92f0eba0
a,0cc175b9c0f1b6a831c399e269772661
1,c4ca4238a0b923820dcc509a6f75849b
,d41d8cd98f00b204e9800998ecf8427e
"""


exasol__models__test_hash_sql = """
with testdata as (

    select * from {{ ref('data_hash') }}

)

select
    {{ hash('input_1') }} as actual,
    output_val as expected

from testdata
"""


exasol__models__test_hash_yml = """
version: 2
models:
  - name: test_hash
    tests:
      - assert_equal:
          actual: actual
          expected: expected
"""

# last_day

exasol__seeds__data_last_day_csv = """date_day,date_part,res
2018-01-02,month,2018-01-31
2018-01-02,quarter,2018-03-31
2018-01-02,year,2018-12-31
,month,
"""


exasol__models__test_last_day_sql = """
with testdata as (

    select * from {{ ref('data_last_day') }}

)

select
    case
        when date_part = 'month' then {{ last_day('date_day', 'month') }}
        when date_part = 'quarter' then {{ last_day('date_day', 'quarter') }}
        when date_part = 'year' then {{ last_day('date_day', 'year') }}
        else null
    end as actual,
    res as expected

from testdata
"""


exasol__models__test_last_day_yml = """
version: 2
models:
  - name: test_last_day
    tests:
      - assert_equal:
          actual: actual
          expected: expected
"""

# length

exasol__seeds__data_length_csv = """expression,output_val
abcdef,6
fishtown,8
december,8
www.google.com/path,19
"""


exasol__models__test_length_sql = """
with testdata as (

    select * from {{ ref('data_length') }}

)

select

    {{ length('expression') }} as actual,
    output_val as expected

from testdata
"""


exasol__models__test_length_yml = """
version: 2
models:
  - name: test_length
    tests:
      - assert_equal:
          actual: actual
          expected: expected
"""

# listagg

exasol__seeds__data_listagg_csv = """group_col,string_text,order_col
1,a,1
1,b,2
1,c,3
2,a,2
2,1,1
2,p,3
3,g,1
3,g,2
3,g,3
"""


exasol__seeds__data_listagg_output_csv = """group_col,expected,version
1,"a_|_b_|_c",bottom_ordered
2,"1_|_a_|_p",bottom_ordered
3,"g_|_g_|_g",bottom_ordered
1,"a_|_b",bottom_ordered_limited
2,"1_|_a",bottom_ordered_limited
3,"g_|_g",bottom_ordered_limited
3,"g, g, g",comma_whitespace_unordered
3,"g",distinct_comma
3,"g,g,g",no_params
"""


exasol__models__test_listagg_sql = """
with testdata as (

    select * from {{ ref('data_listagg') }}

),

data_output as (

    select * from {{ ref('data_listagg_output') }}

),

calculate as (

    select
        group_col,
        {{ listagg('string_text', "'_|_'", "order by order_col") }} as actual,
        'bottom_ordered' as version
    from testdata
    group by group_col

    union all

    select
        group_col,
        {{ listagg('string_text', "'_|_'", "order by order_col", 2) }} as actual,
        'bottom_ordered_limited' as version
    from testdata
    group by group_col

    union all

    select
        group_col,
        {{ listagg('string_text', "', '") }} as actual,
        'comma_whitespace_unordered' as version
    from testdata
    where group_col = 3
    group by group_col

    union all

    select
        group_col,
        {{ listagg('DISTINCT string_text', "','") }} as actual,
        'distinct_comma' as version
    from testdata
    where group_col = 3
    group by group_col

    union all

    select
        group_col,
        {{ listagg('string_text') }} as actual,
        'no_params' as version
    from testdata
    where group_col = 3
    group by group_col

)

select
    calculate.actual,
    data_output.expected
from calculate
left join data_output
on calculate.group_col = data_output.group_col
and calculate.version = data_output.version
"""


exasol__models__test_listagg_yml = """
version: 2
models:
  - name: test_listagg
    tests:
      - assert_equal:
          actual: actual
          expected: expected
"""

# position

exasol__seeds__data_position_csv = """substring_text,string_text,res
def,abcdef,4
land,earth,0
town,fishtown,5
ember,december,4
"""


exasol__models__test_position_sql = """
with testdata as (

    select * from {{ ref('data_position') }}

)

select

    {{ position('substring_text', 'string_text') }} as actual,
    res as expected

from testdata
"""


exasol__models__test_position_yml = """
version: 2
models:
  - name: test_position
    tests:
      - assert_equal:
          actual: actual
          expected: expected
"""

# replace

exasol__seeds__data_replace_csv = """string_text,search_chars,replace_chars,res
a,a,b,b
http://google.com,http://,"",google.com
"""


exasol__models__test_replace_sql = """
with testdata as (

    select

        dr.*,
        coalesce(search_chars, '') as old_chars,
        coalesce(replace_chars, '') as new_chars

    from {{ ref('data_replace') }} dr

)

select

    {{ replace('string_text', 'old_chars', 'new_chars') }} as actual,
    res as expected

from testdata
"""


exasol__models__test_replace_yml = """
version: 2
models:
  - name: test_replace
    tests:
      - assert_equal:
          actual: actual
          expected: expected
"""

# right

exasol__seeds__data_right_csv = """string_text,length_expression,output_val
abcdef,3,def
fishtown,4,town
december,5,ember
december,0,
"""


exasol__models__test_right_sql = """
with testdata as (

    select * from {{ ref('data_right') }}

)

select

    {{ right('string_text', 'length_expression') }} as actual,
    coalesce(output_val, '') as expected

from testdata
"""


exasol__models__test_right_yml = """
version: 2
models:
  - name: test_right
    tests:
      - assert_equal:
          actual: actual
          expected: expected
"""

# safe_cast

exasol__seeds__data_safe_cast_csv = """field,output_val
abc,abc
123,123
,
"""


exasol__models__test_safe_cast_sql = """
with testdata as (

    select * from {{ ref('data_safe_cast') }}

)

select
    {{ safe_cast('field', api.Column.translate_type('string')) }} as actual,
    output_val as expected

from testdata
"""


exasol__models__test_safe_cast_yml = """
version: 2
models:
  - name: test_safe_cast
    tests:
      - assert_equal:
          actual: actual
          expected: expected
"""

# split_part

exasol__seeds__data_split_part_csv = """parts,split_on,result_1,result_2,result_3
a|b|c,|,a,b,c
1|2|3,|,1,2,3
,|,,,
"""


exasol__models__test_split_part_sql = """
with testdata as (

    select * from {{ ref('data_split_part') }}

)

select
    {{ split_part('parts', 'split_on', 1) }} as actual,
    result_1 as expected

from testdata

union all

select
    {{ split_part('parts', 'split_on', 2) }} as actual,
    result_2 as expected

from testdata

union all

select
    {{ split_part('parts', 'split_on', 3) }} as actual,
    result_3 as expected

from testdata
"""


exasol__models__test_split_part_yml = """
version: 2
models:
  - name: test_split_part
    tests:
      - assert_equal:
          actual: actual
          expected: expected
"""



"""
dbt exasol adapter column module
"""
import re
from dataclasses import dataclass
from typing import ClassVar, Dict

from dbt.adapters.base.column import Column
from dbt.exceptions import RuntimeException


@dataclass
# pylint: disable=missing-function-docstring
class ExasolColumn(Column):
    """Column implementation for exasol"""

    # https://docs.exasol.com/db/latest/sql_references/data_types/datatypealiases.htm
    TYPE_LABELS: ClassVar[Dict[str, str]] = {
        "STRING": "VARCHAR(2000000) UTF8",
        "TIMESTAMP": "TIMESTAMP",
        "FLOAT": "DOUBLE PRECISION",
        "INTEGER": "DECIMAL(18,0)",
        "BIGINT": "DECIMAL(36,0)",
        "NUMERIC": "DECIMAL(18,9)",
    }

    def is_numeric(self) -> bool:
        return self.dtype.lower() in ["decimal", "double"]

    def is_integer(self) -> bool:
        # everything that smells like an int is actually a DECIMAL(18, 0)
        return self.is_numeric() and self.numeric_scale == 0

    def is_float(self):
        return self.dtype.lower() == "double"

    def is_string(self) -> bool:
        return self.dtype.lower() in ["char", "varchar"]

    def is_hashtype(self) -> bool:
        return self.dtype.lower() == "hashtype"

    def is_boolean(self) -> bool:
        return self.dtype.lower() == "boolean"

    def is_timestamp(self) -> bool:
        # handle TIMESTAMP and TIMESTAMP WITH LOCAL TIME ZONE
        return self.dtype.lower().startswith("timestamp")

    def is_date(self) -> bool:
        return self.dtype.lower() == "date"

    def string_size(self) -> int:
        if not self.is_string():
            raise RuntimeException("Called string_size() on non-string field!")
        if self.char_size is None:
            return 2000000
        return int(self.char_size)

    @classmethod
    def string_type(cls, size: int) -> str:
        return f"VARCHAR({size})"

    @classmethod
    # pylint: disable=raise-missing-from
    def from_description(cls, name: str, raw_data_type: str) -> "Column":
        match = re.match(r"([^(]+)(\([^)]+\))?", raw_data_type)
        if match is None:
            raise RuntimeException(f'Could not interpret data type "{raw_data_type}"')
        data_type, size_info = match.groups()
        char_size = None
        numeric_precision = None
        numeric_scale = None
        if size_info is not None:
            # strip out the parentheses
            size_info = size_info[1:-1]
            parts = size_info.split(",")
            if len(parts) == 1:
                try:
                    # handle things like HASHTYPE(16 BYTE)
                    size = re.sub(r"[^\d]", "", parts[0])
                    char_size = int(size)
                except ValueError:
                    raise RuntimeException(
                        f'Could not interpret data_type "{raw_data_type}": '
                        f'could not convert "{size}" to an integer'
                    )
            elif len(parts) == 2:
                try:
                    numeric_precision = int(parts[0])
                except ValueError:
                    raise RuntimeException(
                        f'Could not interpret data_type "{raw_data_type}": '
                        f'could not convert "{parts[0]}" to an integer'
                    )
                try:
                    numeric_scale = int(parts[1])
                except ValueError:
                    raise RuntimeException(
                        f'Could not interpret data_type "{raw_data_type}": '
                        f'could not convert "{parts[1]}" to an integer'
                    )

        return cls(name, data_type, char_size, numeric_precision, numeric_scale)



# pylint: disable=wrong-import-order
# pylint: disable=ungrouped-imports
"""
DBT adapter connection implementation for Exasol.
"""
import decimal
import os
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, List, Optional

import agate
import dbt.exceptions
import pyexasol
from dbt.adapters.base import Credentials  # type: ignore
from dbt.adapters.sql import SQLConnectionManager  # type: ignore
from dbt.contracts.connection import AdapterResponse
from dbt.events import AdapterLogger
from hologram.helpers import StrEnum
from pyexasol import ExaConnection

ROW_SEPARATOR_DEFAULT = "LF" if os.linesep == "\n" else "CRLF"
TIMESTAMP_FORMAT_DEFAULT = "YYYY-MM-DDTHH:MI:SS.FF6"

LOGGER = AdapterLogger("exasol")


def connect(**kwargs: bool):
    """
    Global connect method initializing ExasolConnection
    """
    if "autocommit" not in kwargs:
        kwargs["autocommit"] = False
    return ExasolConnection(**kwargs)


class ProtocolVersionType(StrEnum):
    """Exasol protocol versions"""

    V1 = "v1"
    V2 = "v2"
    V3 = "v3"


class ExasolConnection(ExaConnection):
    """
    Override to instantiate ExasolCursor
    """

    row_separator: str = ROW_SEPARATOR_DEFAULT
    timestamp_format: str = TIMESTAMP_FORMAT_DEFAULT

    def cursor(self):
        """Instance of ExasolCursor"""
        return ExasolCursor(self)


@dataclass
class ExasolAdapterResponse(AdapterResponse):
    """
    Override AdapterResponse
    """

    execution_time: Optional[float] = None


# pylint: disable=too-many-instance-attributes
@dataclass
class ExasolCredentials(Credentials):
    """Profile parameters for Exasol in dbt profiles.yml"""

    dsn: str
    user: str
    password: str
    database: str
    schema: str
    # optional statements that can be set in profiles.yml
    # some options might interfere with dbt, so caution is advised
    connection_timeout: int = pyexasol.constant.DEFAULT_CONNECTION_TIMEOUT
    socket_timeout: int = pyexasol.constant.DEFAULT_SOCKET_TIMEOUT
    query_timeout: int = pyexasol.constant.DEFAULT_QUERY_TIMEOUT
    compression: bool = False
    encryption: bool = False
    ## Because of potential interference with dbt,
    # the following statements are not (yet) implemented
    # fetch_dict: bool
    # fetch_size_bytes: int
    # lower_ident: bool
    # quote_ident: bool
    # verbose_error: bool
    # debug: bool
    # udf_output_port: int
    protocol_version: str = "v3"
    retries: int = 1
    row_separator: str = ROW_SEPARATOR_DEFAULT
    timestamp_format: str = TIMESTAMP_FORMAT_DEFAULT

    _ALIASES = {"dbname": "database", "pass": "password"}

    @property
    def type(self):
        return "exasol"

    @property
    def unique_field(self):
        return self.dsn

    def _connection_keys(self):
        return (
            "dsn",
            "user",
            "database",
            "schema",
            "connection_timeout",
            "socket_timeout",
            "query_timeout",
            "compression",
            "encryption",
            "protocol_version",
            "row_separator",
            "timestamp_format",
        )


class ExasolConnectionManager(SQLConnectionManager):
    """Managing Exasol connections"""

    TYPE = "exasol"

    @contextmanager
    def exception_handler(self, sql):
        try:
            yield

        except Exception as yielded_exception:
            LOGGER.debug(f"Error running SQL: {sql}")
            LOGGER.debug("Rolling back transaction.")
            self.rollback_if_open()
            if isinstance(yielded_exception, dbt.exceptions.RuntimeException):
                # during a sql query, an internal to dbt exception was raised.
                # this sounds a lot like a signal handler and probably has
                # useful information, so raise it without modification.
                raise

            raise dbt.exceptions.RuntimeException(yielded_exception)

    @classmethod
    def get_result_from_cursor(cls, cursor: Any) -> agate.Table:
        data: List[Any] = []
        column_names: List[str] = []

        if cursor.description is not None:
            # column_names = [col[0] for col in cursor.description]
            rows = cursor.fetchall()
            for idx, col in enumerate(cursor.description):
                column_names.append(col[0])
                if len(rows) > 0 and isinstance(rows[0][idx], str):
                    if col[1] in ["DECIMAL", "BIGINT"]:
                        for rownum, row in enumerate(rows):
                            tmp = list(row)
                            tmp[idx] = decimal.Decimal(row[idx])
                            rows[rownum] = tmp
            data = cls.process_results(column_names, rows)

        return dbt.clients.agate_helper.table_from_data_flat(data, column_names)  # type: ignore

    @classmethod
    # pylint: disable=raise-missing-from
    def open(cls, connection):
        if connection.state == "open":
            LOGGER.debug("Connection is already open, skipping open.")
            return connection
        credentials = connection.credentials

        # Support protocol versions
        try:
            format_protocol_version = credentials.protocol_version.lower()
            version = ProtocolVersionType(format_protocol_version)

            if version == ProtocolVersionType.V1:
                protocol_version = pyexasol.PROTOCOL_V1
            elif version == ProtocolVersionType.V2:
                protocol_version = pyexasol.PROTOCOL_V2
            else:
                protocol_version = pyexasol.PROTOCOL_V3
        except:
            raise dbt.exceptions.RuntimeException(
                f"{credentials.protocol_version} is not a valid protocol version."
            )

        def _connect():
            conn = connect(
                dsn=credentials.dsn,
                user=credentials.user,
                password=credentials.password,
                autocommit=True,
                connection_timeout=credentials.connection_timeout,
                socket_timeout=credentials.socket_timeout,
                query_timeout=credentials.query_timeout,
                compression=credentials.compression,
                encryption=credentials.encryption,
                protocol_version=protocol_version,
            )
            # exasol adapter specific attributes that are unknown to pyexasol
            # those can be added to ExasolConnection as members
            conn.row_separator = credentials.row_separator
            conn.timestamp_format = credentials.timestamp_format
            conn.execute(
                f"alter session set NLS_TIMESTAMP_FORMAT='{conn.timestamp_format}'"
            )

            return conn

        repeatable_exceptions = [pyexasol.ExaError]

        return cls.retry_connection(
            connection,
            connect=_connect,
            logger=LOGGER,
            retry_limit=credentials.retries,
            retryable_exceptions=repeatable_exceptions,
        )

    def add_begin_query(self):
        return

    def cancel(self, connection):
        connection.abort_query()  # type: ignore

    @classmethod
    def get_response(cls, cursor) -> ExasolAdapterResponse:
        return ExasolAdapterResponse(
            _message="OK",
            rows_affected=cursor.rowcount,
            execution_time=cursor.execution_time,
        )


class ExasolCursor:
    """Exasol dbt-adapter cursor implementation"""

    array_size = 1

    def __init__(self, connection):
        self.connection = connection
        self.stmt = None

    def import_from_file(self, agate_table, table):
        """importing csv skip=1 parameter for header row"""
        self.connection.import_from_file(
            agate_table.original_abspath,
            (table.split(".")[0], table.split(".")[1]),
            import_params={"skip": 1, "row_separator": self.connection.row_separator},
        )
        return self

    def execute(self, query, bindings: Optional[Any] = None):
        """executing query"""
        if query.startswith("0CSV|"):
            self.import_from_file(bindings, query.split("|", 1)[1])  # type: ignore
        else:
            self.stmt = self.connection.execute(query)
        return self

    def fetchone(self):
        """fetch single row"""
        if self.stmt is None:
            raise RuntimeError("Cannot fetch on unset statement")
        return self.stmt.fetchone()

    def fetchmany(self, size=None):
        """fetch single row"""
        if size is None:
            size = self.array_size

        if self.stmt is None:
            raise RuntimeError("Cannot fetch on unset statement")
        return self.stmt.fetchmany(size)

    def fetchall(self):
        """fetch single row"""
        if self.stmt is None:
            raise RuntimeError("Cannot fetch on unset statement")
        return self.stmt.fetchall()

    @property
    def description(self):
        """columns in cursor"""
        cols = []
        if self.stmt is None:
            return cols

        if "resultSet" != self.stmt.result_type:
            return None

        for k, value_set in self.stmt.columns().items():
            cols.append(
                (
                    k,
                    value_set.get("type", None),
                    value_set.get("size", None),
                    value_set.get("size", None),
                    value_set.get("precision", None),
                    value_set.get("scale", None),
                    True,
                )
            )

        return cols

    @property
    def rowcount(self):
        """number of rows in result set"""
        if self.stmt is not None:
            return self.stmt.rowcount()
        return 0

    @property
    def execution_time(self):
        """elapsed time for query"""
        if self.stmt is not None:
            return self.stmt.execution_time
        return 0

    def close(self):
        """closing the cursor / statement"""
        if self.stmt is not None:
            self.stmt.close()



# seeds/my_seed.csv
my_seed_csv = """
ID,NAME,SOME_DATE
1,Easton,1981-05-20T06:46:51
2,Lillian,1978-09-03T18:10:33
3,Jeremiah,1982-03-11T03:59:51
4,Nolan,1976-05-06T20:21:35
""".lstrip()

# models/my_model.sql
my_model_sql = """
select * from {{ ref('my_seed') }}
union all
select null as id, null as name, null as some_date
"""

# models/my_model.yml
my_model_yml = """
version: 2
models:
  - name: my_model
    columns:
      - name: id
        tests:
          - unique
          - not_null  # this test will fail
"""
